#!/usr/bin/perl

use strict;
use warnings;

use File::Basename;

# Migrate a box from Courier to Dovecot
# $0 [ -n | -b | directory ]...
#	-n toggle dry-run (do nothing / do something)
#	-b toggle backup  (create backup of directory prior to make any changes)

my $maildir = '.fb02-mailboxes';	# Default is Maildir

# Defaults for handling stuff
my $dryrun = 1;
my $backup = 1;

# Doit

$| = 1; # Unbuffered

sub handleDir ($) {
	my $dir = shift;

	$dir =~ s!/+$!!;

	unless(-d "$dir/cur") {
		unless(-d "$dir/$maildir/cur") {
			warn "No MailDir: $dir\n" ;
			return;
		}
		$dir .= '/$maildir';
	}

	my ($uid, $gid) = (stat($dir))[4, 5];	# user's IDs
	if($backup && !$dryrun) {
		my $backup = $dir . '.tar.gz';
		if(-e $backup) {
			print "Backup of $dir already exists, skipping\n";
		} else {
			system('tar' , 'cfz', $backup, $dir) == 0
			 or die "Could not make backup for $dir: $?\n";
		}
	}

	# uidlist
	foreach my $uidf (glob("$dir/.*/courierimapuiddb")) {	# YES .* instead of .??*
		my $to = dirname($uidf) . '/dovecot-uidlist';
		if($dryrun) {
			print dirname($uidf) . ": Rename courierimapuiddb to dovecot-uidlist\n";
		} else {
			rename($uidf, $to)
			 or die "Could not rename $uidf to $to $!\n";
		}
	}

	# subscriptions
	if(-r "$dir/courierimapsubscribed") {
		my $from = "$dir/courierimapsubscribed";
		my $to   = "$dir/.subscriptions";
		my $to2  = "$dir/subscriptions";

		if($dryrun) {
			print "$dir: Move courierimapsubscribed to .subscriptions\n"
				. "$dir: Cut \^INBOX\. from .subscriptions\n"
				. "$dir: Link .subscriptions to subscriptions\n";
		} else {
			unless(open(SUB, $from)) {
				die "Could not open $from: $!\n";
			} else {
				unless(open(DOVE, ">$to")) {
					die "Could not open $to: $!\n";
				} else {
					while(<SUB>) {
						s/^INBOX\.//;
						print DOVE;
					}
					close DOVE;
					chmod(0600, $to)
					 or warn "Could not chmod(go=, $to): $!\n";
					chown($uid, $gid, $to)
					 or warn "Could not chown($uid, $gid, $to): $!\n";
					link($to, $to2)	# later versions use this name
					 or warn "Could not link $to to $to2: $!\n";
					unlink($from)
					 or warn "Could not unlink $from: $!\n";
				}
				close SUB;
			}
		}
	}

	# Foldernames
	foreach my $mailbox (glob("$dir/.INBOX.*")) {
		if(-d $mailbox && -d "$mailbox/cur") {	# is maildir mailbox
			my $to = "$dir/" . substr(basename($mailbox), 6);
			if($dryrun) {
				print "$dir: Rename mailbox " . basename($mailbox) . " into "
				 . basename($to) . "\n";
			} else {
				rename($mailbox, $to)
				 or die "Failed to rename $mailbox to $to: $!\n";
			}
		}
	}

	# Contents of .INBOX
	foreach my $mailfile (glob("$dir/.INBOX/{cur,new}/*")) {
		my $to = $dir . substr($mailfile, length($dir) + 7);
		if($dryrun) {
			print "$dir: Move INBOX file "
			 . substr($mailfile, length($dir) + 1)
			 . " to " . substr($to, length($dir) + 1)
			 . "\n"
			 ;
		} else {
			rename($mailfile, $to)
			 or die "Could not move $mailfile to $to $!\n";
		}
	}
	# Delete the useless directories in .INBOX
	foreach my $subdir (glob("$dir/.INBOX/{cur,new,tmp}")) {
		if(-d $subdir) {
			if($dryrun) {
				print "$dir: Remove useless subdir "
				 . basename($subdir) . "\n";
			} else {
				rmdir($subdir)
				 or warn "Failed to rmdir($subdir): $!\n";
			}
		}
	}
	# Propagate the UIDs from .INBOX
	if(-r "$dir/.INBOX/dovecot-uidlist") {
		my $from = "$dir/.INBOX/dovecot-uidlist";
		my $to   = "$dir/dovecot-uidlist";
		if($dryrun) {
			print "$dir: Link UID list of .INBOX to dovecot-uidlist\n";
		} else {
			link($from, $to)
			 or warn "Could not link $from to $to: $!\n";
		}
	}

	# Delete useless maildirfolder (save inodes)
	if(scalar(my @useless = glob("$dir/.*/maildirfolder"))) {	# YES .* instead of .??*
		if($dryrun) {
			print "$dir: Remove useless maildirfolder tags: "
			 . join("\n\t", map { substr($_, length($dir) + 1) } @useless)
			 . "\n";
		} else {
			unlink(@useless);
		}
	}
}

while( my $dir = shift) {
	if($dir eq '-n') {
		$dryrun = !$dryrun;
		next;
	} elsif($dir eq '-b') {
		$backup = !$backup;
		next;
	} elsif($dir =~ /^\-/) {
		print basename($0)
		 . ": [-n | -b | directory]...\n
\t-n: toggle dry-run flag (do nothing / do something) cur:" . ($dryrun? "ON": "OFF") . "
\t-b: toggle backup flag (backup directory before to make any changes,
\t\tThe backup file is directory.tar.gz). cur:" . ($backup? "ON": "OFF") . "
\tdirectory: the Courier Maildir base directory\n\n";
		exit 127;
	}
	print "Migrating $dir"
	 . ($dryrun? " (dry-run, use -n to perform any action)": "")
	 . "\n";
	handleDir($dir);
}

